from math import ceil
from typing import ClassVar, List, Optional, Union

from networkx import Graph
from pydantic import Field

from trano.elements.base import BaseElement
from trano.elements.envelope import (
    BaseExternalWall,
    BaseFloorOnGround,
    BaseInternalElement,
    BaseWindow,
    ExternalWall,
    FloorOnGround,
    InternalElement,
    MergedBaseWall,
    MergedExternalWall,
    MergedWindows,
    WallParameters,
    WindowedWallParameters,
)
from trano.elements.system import BaseOccupancy, Emission, System

MAX_X_SPACES = 3


def _get_controllable_element(elements: List[System]) -> Optional["System"]:
    controllable_elements = []
    for element in elements:
        controllable_ports = element.get_controllable_ports()
        if controllable_ports:
            controllable_elements.append(element)
    if len(controllable_elements) > 1:
        raise NotImplementedError
    if not controllable_elements:
        return None
    return controllable_elements[0]


class Space(BaseElement):
    annotation_template: str = """annotation (
    Placement(transformation(origin = {{ macros.join_list(element.position) }},
    extent = {% raw %}{{-20, -20}, {20, 20}}
    {% endraw %})));"""
    counter: ClassVar[int] = 0
    name: str
    external_boundaries: list[
        Union["BaseExternalWall", "BaseWindow", "BaseFloorOnGround"]
    ]
    internal_elements: List["BaseInternalElement"] = Field(default=[])
    boundaries: Optional[List[WallParameters]] = None
    emissions: List[System] = Field(default=[])
    ventilation_inlets: List[System] = Field(default=[])
    ventilation_outlets: List[System] = Field(default=[])
    occupancy: Optional[BaseOccupancy] = None

    def model_post_init(self, __context) -> None:  # type: ignore # noqa: ANN001
        self._assign_space()

    def _assign_space(self) -> None:
        for emission in (
            self.emissions + self.ventilation_inlets + self.ventilation_outlets
        ):
            if emission.control:
                emission.control.space_name = self.name
        if self.occupancy:
            self.occupancy.space_name = self.name

    @property
    def number_merged_external_boundaries(self) -> int:
        return sum(
            [
                boundary.length
                for boundary in self.merged_external_boundaries + self.internal_elements
            ]
        )

    @property
    def number_ventilation_ports(self) -> int:
        return 2 + 1  # databus

    @property
    def merged_external_boundaries(
        self,
    ) -> List[
        Union["BaseExternalWall", "BaseWindow", "BaseFloorOnGround", "MergedBaseWall"]
    ]:

        external_walls = [
            boundary
            for boundary in self.external_boundaries
            if boundary.type in ["ExternalWall", "ExternalDoor"]
        ]
        windows = [
            boundary
            for boundary in self.external_boundaries
            if boundary.type == "Window"
        ]
        merged_external_walls = MergedExternalWall.from_base_elements(external_walls)
        merged_windows = MergedWindows.from_base_windows(windows)  # type: ignore
        external_boundaries: list[
            BaseExternalWall | BaseWindow | BaseFloorOnGround | MergedBaseWall
        ] = (
            merged_external_walls
            + merged_windows
            + [
                boundary
                for boundary in self.external_boundaries
                if boundary.type not in ["ExternalWall", "Window", "ExternalDoor"]
            ]
        )
        return external_boundaries

    def get_controllable_emission(self) -> Optional["System"]:
        return _get_controllable_element(self.emissions)

    def assign_position(self) -> None:
        self.position = [
            250 * (Space.counter % MAX_X_SPACES),
            150 * ceil(Space.counter / MAX_X_SPACES),
        ]
        Space.counter += 1
        x = self.position[0]
        y = self.position[1]
        for i, emission in enumerate(self.emissions):
            emission.position = [x + i * 30, y - 75]
        if self.occupancy:
            self.occupancy.position = [x - 50, y]

    def find_emission(self) -> Optional["Emission"]:
        emissions = [
            emission for emission in self.emissions if isinstance(emission, Emission)
        ]
        if not emissions:
            return None
        if len(emissions) != 1:
            raise NotImplementedError
        return emissions[0]

    def first_emission(self) -> Optional["System"]:
        if self.emissions:
            return self.emissions[0]
        return None

    def last_emission(self) -> Optional["System"]:
        if self.emissions:
            return self.emissions[-1]
        return None

    def get_ventilation_inlet(self) -> Optional["System"]:
        if self.ventilation_inlets:
            return self.ventilation_inlets[-1]
        return None

    def get_last_ventilation_inlet(self) -> Optional["System"]:
        if self.ventilation_inlets:
            return self.ventilation_inlets[0]
        return None

    def get_ventilation_outlet(self) -> Optional["System"]:
        if self.ventilation_outlets:
            return self.ventilation_outlets[0]
        return None

    def get_last_ventilation_outlet(self) -> Optional["System"]:
        if self.ventilation_outlets:
            return self.ventilation_outlets[-1]
        return None

    def get_neighhors(self, graph: Graph) -> None:

        neighbors = list(graph.neighbors(self))  # type: ignore
        self.boundaries = []
        windowed_wall_parameters = WindowedWallParameters.from_neighbors(neighbors)
        for wall in [ExternalWall, BaseWindow, InternalElement, FloorOnGround]:
            self.boundaries.append(
                WallParameters.from_neighbors(
                    neighbors,
                    wall,  # type: ignore
                    filter=windowed_wall_parameters.included_external_walls,
                )
            )
        self.boundaries += [windowed_wall_parameters]

    def __add__(self, other: "Space") -> "Space":
        self.name = (
            f"merge_{self.name.replace('merge', '')}_{other.name.replace('merge', '')}"
        )
        self.volume: float = self.volume + other.volume
        self.external_boundaries += other.external_boundaries
        return self



from enum import Enum
from typing import List, Literal, Optional

from pydantic import BaseModel, Field


class Tilt(Enum):
    wall = "wall"
    ceiling = "ceiling"
    floor = "floor"


class Azimuth:
    north = 0
    south = 90
    east = 45
    west = 135


class Flow(Enum):
    inlet = "inlet"
    outlet = "outlet"
    inlet_or_outlet = "inlet_or_outlet"
    undirected = "undirected"
    interchangeable_port = "interchangeable_port"


Boolean = Literal["true", "false"]


class Line(BaseModel):
    template: str
    key: Optional[str] = None
    color: str = "grey"
    label: str
    line_style: str = "solid"
    line_width: float = 1.5


class Axis(BaseModel):
    lines: List[Line] = Field(default=[])
    label: str


class PartialConnection(BaseModel):
    equation: str
    position: List[float]


class ConnectionView(BaseModel):
    color: Optional[str] = "{255,204,51}"
    thickness: float = 0.5


class BaseVariant:
    default: str = "default"


DynamicTemplateCategories = Literal["ventilation", "control", "fluid", "boiler"]



import copy
import json
import tempfile
from collections import Counter
from functools import partial
from pathlib import Path
from typing import Any, Dict, List, Optional

import yaml
from linkml.validator import validate_file  # type: ignore
from pydantic import BaseModel

from trano.data.include import Loader
from trano.data_models.converter import converter
from trano.elements import (
    Boundary,
    ExternalWall,
    FloorOnGround,
    InternalElement,
    Window,
    param_from_config,
)
from trano.elements.construction import (
    Construction,
    Gas,
    GasLayer,
    Glass,
    GlassLayer,
    GlassMaterial,
    Layer,
    Material,
)
from trano.elements.control import BoilerControl  # noqa: F401
from trano.elements.space import Space

# TODO: fix these imports
from trano.elements.system import Boiler  # noqa: F401
from trano.elements.system import AirHandlingUnit, Occupancy, Weather
from trano.elements.types import Tilt
from trano.elements.utils import import_element_function
from trano.library.library import Library
from trano.topology import Network

SpaceParameter = param_from_config("Space")
DATA_MODEL_PATH = Path(__file__).parent.joinpath("trano_final.yaml")
COUNTER: Dict[Any, Any] = Counter()


def to_camel_case(snake_str: str) -> str:
    return "".join(x.capitalize() for x in snake_str.lower().split("_"))


class Component(BaseModel):
    name: str
    component_instance: Any


def validate_model(data: Any, suffix: str) -> None:  # noqa: ANN401
    with tempfile.NamedTemporaryFile(mode="w+", suffix=suffix, delete=False) as f2:
        enriched_path = Path(f2.name)
        enriched_path.write_text(json.dumps(data))
        report = validate_file(enriched_path, DATA_MODEL_PATH, "Building")
        if report.results:
            raise Exception("Invalid model.")


def _instantiate_component(component_: Dict[str, Any]) -> Component:
    component = copy.deepcopy(component_)
    components = component.items()
    if len(components) != 1:
        raise NotImplementedError("Only one component type is allowed")
    component_type, component_parameters = next(iter(components))
    component_parameters.pop("inlets", None)
    component_parameters.pop("outlets", None)
    component_type = to_camel_case(component_type)
    # TODO: just find a way to import the required components directly here!

    component_class = import_element_function(component_type)
    name = component_parameters.pop("id")
    component_parameters.update({"name": name})
    component_parameters_class = param_from_config(component_type)
    if "parameters" in component_parameters and component_parameters_class is not None:
        parameters = component_parameters_class(**component_parameters["parameters"])
        component_parameters = component_parameters | {"parameters": parameters}
    if component_parameters.get("control"):
        controls = component_parameters["control"].items()
        if len(controls) != 1:
            raise NotImplementedError("Only one component type is allowed")
        control_type, control_parameter = next(iter(controls))
        control_class = import_element_function(to_camel_case(control_type))
        control_name = control_parameter.pop("id", None)
        if control_name:
            control_parameter.update({"name": control_name})
        component_parameter_class = param_from_config(f"{to_camel_case(control_type)}")
        if "parameters" in control_parameter and component_parameter_class is not None:
            parameters = component_parameter_class(**control_parameter["parameters"])
            control_parameter = control_parameter | {"parameters": parameters}
        control = control_class(**control_parameter)
        component_parameters.update({"control": control})
    component = component_class(**component_parameters)
    return Component(name=name, component_instance=component)


class EnrichedModel(BaseModel):
    data: Any


def load_and_enrich_model(model_path: Path) -> EnrichedModel:
    if model_path.suffix == ".yaml":
        load_function = partial(yaml.load, Loader=Loader)
        dump_function = yaml.safe_dump
    elif model_path.suffix == ".json":
        load_function = json.loads  # type: ignore
        dump_function = json.dump  # type: ignore # TODO: why?
    else:
        raise Exception("Invalid file format")
    data = load_function(model_path.read_text())
    data = assign_space_id(data)
    _parse(data)
    with tempfile.NamedTemporaryFile(
        mode="w+", suffix=model_path.suffix, delete=False  # TODO: delete later?
    ) as f:
        dump_function(data, f)

    converted_data = converter(
        input=str(Path(f.name)),
        target_class="Building",
        schema=str(DATA_MODEL_PATH),
        output_format=model_path.suffix[1:],
    )
    return EnrichedModel(data=converted_data)


def _build_materials(data: Dict[str, Any]) -> Dict[str, Any]:
    materials = {}
    material_types = {"material": Material, "gas": Gas, "glass_material": GlassMaterial}
    for material_type, material_class in material_types.items():
        for material in data.get(material_type, []):
            materials[material["id"]] = material_class(
                **(material | {"name": material["id"]})
            )
    return materials


def _merge_default(data: Dict[str, Any]) -> Dict[str, Any]:
    data["constructions"] = data.get("constructions", []) + data.get("default", {}).get(
        "constructions", []
    )
    data["material"] = data.get("material", []) + data.get("default", {}).get(
        "material", []
    )
    data["glazings"] = data.get("glazings", []) + data.get("default", {}).get(
        "glazings", []
    )
    data["gas"] = data.get("gas", []) + data.get("default", {}).get("gas", [])
    data["glass_material"] = data.get("glass_material", []) + data.get(
        "default", {}
    ).get("glass_material", [])
    data.pop("default", None)
    return data


# TODO: reduce complexity
def convert_network(  # noqa: PLR0915, C901, PLR0912
    name: str, model_path: Path, library: Optional[Library] = None
) -> Network:
    network = Network(name=name, library=library)
    occupancy = None
    system_counter: Any = Counter()

    enriched_model = load_and_enrich_model(model_path)
    validate_model(enriched_model.data, model_path.suffix)
    data = enriched_model.data
    data = _merge_default(data)
    materials = _build_materials(data)
    constructions: Dict[str, Construction | Glass] = {}
    for construction in data["constructions"]:
        layers = [
            Layer(**(layer | {"material": materials[layer["material"]]}))
            for layer in construction["layers"]
        ]
        constructions[construction["id"]] = Construction(
            name=construction["id"], layers=layers
        )
    for glazing in data.get("glazings", []):
        glazing_layers: List[GasLayer | GlassLayer] = []
        for layer in glazing["layers"]:
            if layer.get("gas", None):
                glazing_layers.append(
                    GasLayer(
                        thickness=layer["thickness"], material=materials[layer["gas"]]
                    )
                )
            if layer.get("glass", None):
                glazing_layers.append(
                    GlassLayer(
                        thickness=layer["thickness"], material=materials[layer["glass"]]
                    )
                )
        constructions[glazing["id"]] = Glass(
            name=glazing["id"],
            layers=glazing_layers,
            u_value_frame=glazing["u_value_frame"],
        )

    spaces = []
    space_dict = {}
    systems = {}
    for space in data["spaces"]:
        external_boundaries = space["external_boundaries"]
        external_walls: List[ExternalWall | Window | FloorOnGround] = []
        for external_wall in external_boundaries.get("external_walls", []):
            external_wall_ = ExternalWall(
                **(
                    external_wall
                    | {"construction": constructions[external_wall["construction"]]}
                )
            )
            external_walls.append(external_wall_)
        for window in external_boundaries.get("windows", []):
            window_ = Window(
                **(window | {"construction": constructions[window["construction"]]})
            )
            external_walls.append(window_)
        for floor_on_ground in external_boundaries.get("floor_on_grounds", []):
            floor_on_ground_ = FloorOnGround(
                **(
                    floor_on_ground
                    | {"construction": constructions[floor_on_ground["construction"]]}
                )
            )
            external_walls.append(floor_on_ground_)
        occupancy_parameter_class = param_from_config("Occupancy")
        if space.get("occupancy") is not None and occupancy_parameter_class is not None:
            system_counter.update(["occupancy"])

            occupancy_ = space["occupancy"]
            occupancy = Occupancy(
                **(
                    occupancy_
                    | {"name": f"occupancy_{system_counter['occupancy']}"}
                    | {
                        "parameters": occupancy_parameter_class(
                            **occupancy_.get("parameters", {})
                        )
                    }
                )
            )
        elif (
            network.library.default_parameters.get("occupancy") is not None
            and occupancy_parameter_class is not None
        ):
            system_counter.update(["occupancy"])
            occupancy = Occupancy(
                name=f"occupancy_{system_counter['occupancy']}",
                parameters=occupancy_parameter_class(
                    **network.library.default_parameters["occupancy"]
                ),
            )
        emissions = []
        for emission in space.get("emissions", []):
            emission_ = _instantiate_component(emission)
            systems[emission_.name] = emission_.component_instance
            emissions.append(emission_.component_instance)
        ventilation_inlets = []
        for inlet in space.get("ventilation_inlets", []):
            inlet_ = _instantiate_component(inlet)
            systems[inlet_.name] = inlet_.component_instance
            ventilation_inlets.append(inlet_.component_instance)
        ventilation_outlets = []
        for outlet in space.get("ventilation_outlets", []):
            outlet_ = _instantiate_component(outlet)
            systems[outlet_.name] = outlet_.component_instance
            ventilation_outlets.append(outlet_.component_instance)
        if SpaceParameter is None:
            raise Exception("SpaceParameter is not defined")
        space_ = Space(
            name=space["id"],
            external_boundaries=external_walls,
            occupancy=occupancy,
            parameters=SpaceParameter(**space["parameters"]),
            emissions=emissions,
            ventilation_inlets=ventilation_inlets,
            ventilation_outlets=ventilation_outlets,
        )
        space_dict[space["id"]] = space_
        spaces.append(space_)
    create_internal = not data.get("internal_walls", [])

    if (
        data.get("weather", {}).get("parameters")
        and param_from_config("Weather") is not None
    ):
        weather = Weather(
            name="weather",
            parameters=param_from_config("Weather")(**data["weather"]["parameters"]),  # type: ignore
        )
    else:
        weather = Weather(name="weather")

    network.add_boiler_plate_spaces(
        spaces, weather=weather, create_internal=create_internal
    )
    for internal_wall in data.get("internal_walls", []):
        space_1 = space_dict[internal_wall["space_1"]]
        space_2 = space_dict[internal_wall["space_2"]]
        internal_element = InternalElement(
            name=f"internal_{space_1.name}_{space_2.name}_{internal_wall['construction'].lower().split(':')[0]}",
            surface=internal_wall["surface"],
            azimuth=10,
            construction=constructions[internal_wall["construction"]],
            tilt=Tilt.wall,
        )
        network.connect_spaces(space_1, space_2, internal_element=internal_element)

    edges = []
    for system in data["systems"]:
        system_ = _instantiate_component(system)
        systems[system_.name] = system_.component_instance
    for system in data["systems"]:
        for value in system.values():
            edges += [
                (systems[value["id"]], systems[outlet])
                for outlet in value.get("outlets", [])
            ]
            edges += [
                (systems[inlet], systems[value["id"]])
                for inlet in value.get("inlets", [])
            ]
    for edge in edges:
        network.connect_systems(*edge)

    ahus = [n for n in network.graph.nodes if isinstance(n, AirHandlingUnit)]
    if ahus:
        boundary = Boundary(name="boundary")
        network.connect_elements(boundary, ahus[0])
        weather = next(n for n in network.graph.nodes if isinstance(n, Weather))
        network.connect_elements(boundary, weather)
    return network


def convert_model(name: str, model_path: Path) -> str:
    network = convert_network(name, model_path)
    return network.model()


def _parse(data: Dict[str, Any]) -> None:
    for k, v in data.items():
        if isinstance(v, dict):
            _parse(v)
        elif isinstance(v, list):
            for i in v:
                if isinstance(i, dict):
                    _parse(i)
        elif v is None:
            if "control" in k:
                COUNTER.update(["control"])  # type: ignore
                data[k] = {"id": f"CONTROL:{COUNTER['control']}"}
            if "occupancy" in k:
                COUNTER.update(["occupancy"])  # type: ignore
                data[k] = {"parameters": {"occupancy": "3600 * {7, 19}"}}


def assign_space_id(data: Dict[str, Any]) -> Dict[str, Any]:
    space_counter: Dict[Any, Any] = Counter()
    spaces = []
    for space in data.get("spaces", []):
        space_counter.update(["space"])  # type: ignore
        spaces.append({"id": f"SPACE:{space_counter['space']}"} | space)
    data["spaces"] = spaces
    return data



from pathlib import Path
from typing import List, Optional

import pandas as pd
from pydantic import BaseModel, Field, computed_field

from trano.elements.base import BaseElement
from trano.elements.inputs import BaseInputOutput


class ValidationData(BaseModel):
    data: Optional[str] = None
    columns: List[str] = Field([])


class DataBus(BaseElement):
    name: str
    position: Optional[List[float]] = None
    spaces: List[str]
    non_connected_ports: List[BaseInputOutput] = Field(default=[])
    external_data: Optional[Path] = None

    @computed_field
    def validation_data(self) -> ValidationData:
        if not self.external_data:
            return ValidationData()
        return transform_csv_to_table(self.external_data)


def transform_csv_to_table(
    file_path: Path, total_second: bool = True
) -> ValidationData:
    data = pd.read_csv(file_path, index_col=0, infer_datetime_format=True)
    data = data.ffill().bfill()
    data = data.dropna(axis=1)
    data.index = pd.to_datetime(data.index)
    if total_second:
        data.index = (data.index - data.first_valid_index()).total_seconds()  # type: ignore
    else:
        data.index = data.index.astype(int) // 10**9
    data_str = data.to_csv(sep=",", header=False, lineterminator=";")
    if data_str.endswith(";"):
        data_str = data_str[:-1]
    return ValidationData(data=data_str, columns=data.columns.tolist())



import json
import logging
from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Union

from pydantic import BaseModel, Field

from trano.elements.inputs import (
    BaseInputOutput,
    BooleanInput,
    BooleanOutput,
    IntegerInput,
    IntegerOutput,
    RealInput,
    RealOutput,
    Target,
)
from trano.exceptions import ControllerBusPortError

logger = logging.getLogger(__name__)
if TYPE_CHECKING:
    from trano.elements import BaseElement


def _evaluate(
    element: "BaseElement", commands: List[str]
) -> Optional[Union[str, "BaseElement", List["BaseElement"], List[str]]]:
    for command in commands:
        if hasattr(element, command):
            element = getattr(element, command)
        else:
            raise Exception(f" Element {element.name} has no command {command}")
    return element


def _evaluate_target(target: Target, element: "BaseElement") -> str | List[str]:
    from trano.elements import BaseElement

    target_ = _evaluate(element, target.commands())
    if not target_:
        message = (
            f"Element {element.name} of type {type(element).__name__} "
            f"has no valid target {target.commands()}. "
            f"This indicates that the selected element is incompatible."
        )
        logger.error(message)
        raise ControllerBusPortError(message)
    if isinstance(target_, list) and all(isinstance(t, BaseElement) for t in target_):
        return [
            _evaluate(sub, target.sub_commands())  # type: ignore
            for sub in target_
            if isinstance(sub, BaseElement)
        ]
    return target_  # type: ignore


def _append_to_port(
    input_: BaseInputOutput,
    ports: Dict[str, List[BaseInputOutput]],
    target: Target,
    evaluated_element: str,
    element: "BaseElement",
) -> Dict[str, List[BaseInputOutput]]:
    ports[type(input_).__name__].append(
        type(input_)(
            **(
                input_.model_dump()
                | {
                    "target": Target(
                        **(
                            target.model_dump()
                            | {
                                "evaluated_element": element.name,
                                "value": evaluated_element,
                            }
                        )
                    ),
                }
            )
        )
    )
    return ports


class ControllerBus(BaseModel):
    template: str = """Controls.BaseClasses.DataBus dataBus
    annotation (Placement(transformation(
  extent={{-120,-18},{-80,22}}), iconTransformation(extent={{-120,62},{-78,98}})));"""
    real_inputs: list[RealInput] = Field(default=[])
    real_outputs: list[RealOutput] = Field(default=[])
    integer_inputs: list[IntegerInput] = Field(default=[])
    integer_outputs: list[IntegerOutput] = Field(default=[])
    boolean_inputs: list[BooleanInput] = Field(default=[])
    boolean_outputs: list[BooleanOutput] = Field(default=[])

    @classmethod
    def from_configuration(cls, file_path: Path) -> "ControllerBus":
        return cls(**json.loads(file_path.read_text()))

    def inputs(
        self,
    ) -> List[
        BooleanInput
        | IntegerOutput
        | IntegerInput
        | RealOutput
        | RealInput
        | BooleanOutput
    ]:
        return (
            self.real_inputs
            + self.real_outputs
            + self.integer_inputs
            + self.integer_outputs
            + self.boolean_inputs
            + self.boolean_outputs
        )

    def _get_targets(self) -> Dict[Target, List[BaseInputOutput]]:
        return {
            input.target: [
                input_ for input_ in self.inputs() if input_.target == input.target
            ]
            for input in self.inputs()
        }

    def list_ports(
        self, element: "BaseElement", **kwargs: Any  # noqa: ANN401
    ) -> Dict[str, List[BaseInputOutput]]:
        ports: Dict[str, List[BaseInputOutput]] = {
            "RealOutput": [],
            "RealInput": [],
            "IntegerOutput": [],
            "IntegerInput": [],
            "BooleanOutput": [],
            "BooleanInput": [],
        }
        for target, inputs in self._get_targets().items():
            # TODO: Fix this
            evaluated_element = _evaluate_target(target, element)
            for input in inputs:
                if isinstance(evaluated_element, list):
                    for evaluated_element_ in evaluated_element:
                        ports = _append_to_port(
                            input, ports, target, evaluated_element_, element
                        )
                else:
                    ports = _append_to_port(
                        input, ports, target, evaluated_element, element
                    )
        return ports

    def bus_ports(
        self, element: "BaseElement", **kwargs: Any  # noqa: ANN401
    ) -> List[str]:
        ports: List[str] = []
        for target, inputs in self._get_targets().items():
            target_value = _evaluate_target(target, element)
            for input in inputs:
                if isinstance(target_value, list):
                    for i, target_ in enumerate(target_value):
                        ports = _append_ports(
                            ports,
                            input,
                            target_,
                            "multi",
                            f".{input.port}[{i + 1}]",
                            f"[{i + 1}].{input.port}",
                        )
                else:  # : PLR5501
                    ports = _append_ports(
                        ports,
                        input,
                        target_value,
                        "port",
                        f".{input.port}",
                        "",
                    )
        return ports


def _append_ports(  # noqa: PLR0913
    ports: List[str],
    input: BaseInputOutput,
    evaluated_target: str,
    case_1_condition: str,
    case_1: str,
    case_2: str,
) -> List[str]:
    if getattr(input, case_1_condition):

        ports.append(
            f"connect(dataBus.{input.name}{evaluated_target.capitalize()}, "
            f"{input.component}{case_1});"
        )
    else:
        ports.append(
            f"connect(dataBus.{input.name}{evaluated_target.capitalize()}, "
            f"{input.component}{case_2});"
        )
    return ports



from typing import Any, Callable, ClassVar, Dict, List, Optional

from pydantic import BaseModel, ConfigDict, Field, field_validator, model_validator

from trano.elements.components import DynamicComponentTemplate
from trano.elements.connection import Port, _has_inlet_or_outlet, _is_inlet_or_outlet
from trano.elements.figure import NamedFigure
from trano.elements.parameters import BaseParameter, param_from_config
from trano.elements.types import BaseVariant, Flow
from trano.library.library import AvailableLibraries, Library


class PortFound(Exception):  # noqa: N818
    def __init__(self, message: str, port: Port) -> None:
        super().__init__(message)
        self.port = port


def _unique_port(ports: List[Port]) -> None:
    if ports and len(ports) != 1:
        raise NotImplementedError

    if ports:
        raise PortFound("Port found", port=ports[0])


def _one_port_max(ports: List[Port]) -> None:
    if ports:
        if len(ports) > 1:
            raise NotImplementedError
        raise PortFound("Port found", port=ports[0])


def _ports_exist(ports: List[Port]) -> None:
    if ports:
        raise PortFound("Port found", port=ports[0])


def _available_ports_with_targets(
    self_ports: List[Port], target: "BaseElement"
) -> List[Port]:
    return [
        port
        for port in self_ports
        if port.targets
        and any(isinstance(target, target_) for target_ in port.targets)
        and port.is_available()
    ]


def _available_ports_without_targets(self_ports: List[Port]) -> List[Port]:
    return [port for port in self_ports if not port.targets and port.is_available()]


def _ports_with_specified_flow(_available_ports: List[Port], flow: Flow) -> List[Port]:
    return [port for port in _available_ports if port.flow == flow]


def _apply_function_to_inlet_outlet_ports(
    _available_ports: List[Port],
    target: "BaseElement",
    function: Callable[["BaseElement"], bool],
) -> List[Port]:
    return [
        port
        for port in _ports_with_specified_flow(_available_ports, Flow.inlet_or_outlet)
        if function(target)
    ]


class BaseElement(BaseModel):
    name_counter: ClassVar[int] = (
        0  # TODO: this needs to be removed and replaced with a proper solution.
    )
    name: Optional[str] = Field(default=None)
    annotation_template: str = """annotation (
    Placement(transformation(origin = {{ macros.join_list(element.position) }},
    extent = {% raw %}{{-10, -10}, {10, 10}}
    {% endraw %})));"""
    model_config = ConfigDict(arbitrary_types_allowed=True, extra="allow")
    parameters: Optional[BaseParameter] = None
    position: Optional[List[float]] = None
    ports: list[Port] = Field(default=[], validate_default=True)
    template: Optional[str] = None
    component_template: Optional[DynamicComponentTemplate] = None
    variant: str = BaseVariant.default
    libraries_data: Optional[AvailableLibraries] = None
    figures: List[NamedFigure] = Field(default=[])

    @model_validator(mode="before")
    @classmethod
    def validate_libraries_data(cls, value: Dict[str, Any]) -> Dict[str, Any]:
        libraries_data = AvailableLibraries.from_config(cls.__name__)
        if libraries_data:
            value["libraries_data"] = libraries_data

        parameter_class = param_from_config(cls.__name__)
        if parameter_class and isinstance(value, dict) and not value.get("parameters"):
            value["parameters"] = parameter_class()

        return value

    @model_validator(mode="after")
    def assign_default_name(self) -> "BaseElement":
        if self.name is None:
            self.name = f"{type(self).__name__.lower()}_{type(self).name_counter}"
            type(self).name_counter += 1
        return self

    @field_validator("name")
    @classmethod
    def clean_name(cls, value: str) -> str:
        if ":" in value:
            return value.lower().replace(":", "_")
        return value

    def assign_library_property(self, library: "Library") -> bool:
        if not self.libraries_data:
            return False
        library_data = self.libraries_data.get_library_data(library, self.variant)
        if not library_data:
            return False
        if not self.ports:
            self.ports = library_data.ports_factory()
        if not self.template:
            self.template = library_data.template
        if not self.component_template:
            self.component_template = library_data.component_template
        if not self.figures and library_data.figures:
            self.figures = [
                NamedFigure(**(fig.render_key(self).model_dump() | {"name": self.name}))
                for fig in library_data.figures
            ]

        return True

    def processed_parameters(self, library: "Library") -> Any:  # noqa: ANN401
        if self.libraries_data:
            library_data = self.libraries_data.get_library_data(library, self.variant)
            if library_data and self.parameters:
                return library_data.parameter_processing(self.parameters)
        return {}

    def get_position(self, layout: Dict["BaseElement", Any]) -> None:
        if not self.position:
            self.position = list(layout.get(self))  # type: ignore

    def get_controllable_ports(self) -> List[Port]:
        return [port for port in self.ports if port.is_controllable()]

    @property
    def type(self) -> str:
        return type(self).__name__

    def _ports(self, ports_to_skip: List[Port]) -> List[Port]:
        return [port for port in self.ports if port not in ports_to_skip]

    def _get_target_compatible_port(
        self, target: "BaseElement", flow: Flow, ports_to_skip: List[Port]
    ) -> Optional["Port"]:
        self_ports = self._ports(ports_to_skip)
        available_ports = _available_ports_with_targets(self_ports, target)
        available_ports_without_target = _available_ports_without_targets(self_ports)
        available_ports_with_interchangeable_flow = _ports_with_specified_flow(
            available_ports, Flow.interchangeable_port
        )
        available_ports_with_undirected_flow = _ports_with_specified_flow(
            available_ports, Flow.undirected
        )
        available_ports_with_directed_flow = _ports_with_specified_flow(
            available_ports, flow
        )
        available_ports_with_inlet_or_outlet = _apply_function_to_inlet_outlet_ports(
            available_ports, target, _is_inlet_or_outlet
        )
        available_ports_without_target_with_inlet_outlet = (
            _apply_function_to_inlet_outlet_ports(
                available_ports_without_target, target, _has_inlet_or_outlet
            )
        )
        available_ports_without_target_with_directed_flow = _ports_with_specified_flow(
            available_ports_without_target, flow
        )
        try:
            _ports_exist(available_ports_with_interchangeable_flow)
            _unique_port(available_ports_with_undirected_flow)
            _ports_exist(available_ports_with_inlet_or_outlet)
            _unique_port(available_ports_with_directed_flow)
            _one_port_max(available_ports_without_target_with_inlet_outlet)
            _one_port_max(available_ports_without_target_with_directed_flow)
        except PortFound as port:
            return port.port
        except NotImplementedError:
            raise
        except Exception:
            raise
        return None

    def __hash__(self) -> int:
        return hash(f"{self.name}-{type(self).__name__}")


# This has to be here for now!!!
class Control(BaseElement):
    position: Optional[List[float]] = None
    controllable_element: Optional[BaseElement] = None
    space_name: Optional[str] = None



import logging
from typing import Optional

import docker  # type: ignore

from trano.simulate.simulate import SimulationOptions

logger = logging.getLogger(__name__)


def is_success(
    results: docker.models.containers.ExecResult,
    options: Optional[SimulationOptions] = None,
) -> bool:
    logger.debug(f"Simulation results: {results.output.decode()}")
    if options and options.check_only:
        return "true" in results.output.decode()
    return "The simulation finished successfully" in results.output.decode()



import itertools
import shutil
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple, Type, Union

import matplotlib.pyplot as plt
import networkx as nx
import numpy as np
from jinja2 import Environment, FileSystemLoader
from networkx import DiGraph, shortest_path
from networkx.classes.reportviews import NodeView
from pyvis.network import Network as PyvisNetwork  # type: ignore

from tests.constructions.constructions import Constructions
from trano.elements import (
    BaseElement,
    Connection,
    Control,
    DynamicTemplateCategories,
    InternalElement,
    connect,
)
from trano.elements.bus import DataBus
from trano.elements.construction import extract_properties
from trano.elements.control import AhuControl, CollectorControl, VAVControl
from trano.elements.inputs import BaseInputOutput
from trano.elements.space import Space, _get_controllable_element
from trano.elements.system import (
    VAV,
    AirHandlingUnit,
    Boiler,
    Pump,
    System,
    TemperatureSensor,
    ThreeWayValve,
    Valve,
    Ventilation,
    Weather,
)
from trano.elements.types import Tilt
from trano.exceptions import WrongSystemFlowError
from trano.library.library import Library


class Network:  # : PLR0904, #TODO: fix this
    def __init__(
        self,
        name: str,
        library: Optional[Library] = None,
        external_data: Optional[Path] = None,
    ) -> None:
        self.graph: DiGraph = DiGraph()
        self.edge_attributes: List[Connection] = []
        self.name: str = name
        self._system_controls: List[Control] = []
        self.library = library or Library.load_default()
        self.external_data = external_data
        self.dynamic_components: Dict[DynamicTemplateCategories, List[str]] = {
            "ventilation": [],
            "control": [],
            "boiler": [],
        }

    def add_node(self, node: BaseElement) -> None:

        if not node.libraries_data:
            return
            # TODO: check better option here!!
        found_library = node.assign_library_property(self.library)
        if not found_library:
            return

        if node not in self.graph.nodes:
            self.graph.add_node(node)
        if (isinstance(node, System) and node.control) and (
            node.control not in self.graph.nodes
        ):
            node_control = node.control
            if not node_control.libraries_data:
                raise Exception(
                    f"No library data defined for NOde of type {type(node).__name__}"
                )
            node_control.assign_library_property(self.library)
            self.graph.add_node(node_control)
            self.graph.add_edge(node, node_control)
            node_control.controllable_element = node

    def add_space(self, space: "Space") -> None:
        self.add_node(space)
        if self.library.merged_external_boundaries:
            external_boundaries = space.merged_external_boundaries
        else:
            external_boundaries = space.external_boundaries  # type: ignore
        for boundary in external_boundaries:
            self.add_node(boundary)
            self.graph.add_edge(
                space,
                boundary,
            )
        self._build_space_emission(space)  # TODO: perhaps move to space
        self._build_occupancy(space)
        self._build_space_ventilation(space)
        space.assign_position()

    def _add_subsequent_systems(self, systems: List[System]) -> None:
        for system1, system2 in zip(systems[:-1], systems[1:]):
            if not self.graph.has_node(system1):  # type: ignore
                self.add_node(system1)
            if not self.graph.has_node(system2):  # type: ignore
                self.add_node(system2)
            self.graph.add_edge(
                system1,
                system2,
            )

    def _build_space_ventilation(self, space: "Space") -> None:
        # Assumption: first element always the one connected to the space.
        if space.get_ventilation_inlet():
            self.add_node(space.get_ventilation_inlet())  # type: ignore
            self.graph.add_edge(space.get_ventilation_inlet(), space)
        if space.get_ventilation_outlet():
            self.add_node(space.get_ventilation_outlet())  # type: ignore
            self.graph.add_edge(space, space.get_ventilation_outlet())
        # The rest is connected to each other
        self._add_subsequent_systems(space.ventilation_outlets)
        self._add_subsequent_systems(space.ventilation_inlets)

    def _build_space_emission(self, space: "Space") -> None:
        emission = space.find_emission()
        if emission:
            self.add_node(emission)
            self.graph.add_edge(
                space,
                emission,
            )
            self._add_subsequent_systems(space.emissions)

    def _build_data_bus(self) -> DataBus:
        # TODO: this feels like it does not belong here!!!!

        spaces = sorted(
            [node for node in self.graph.nodes if isinstance(node, Space)],
            key=lambda x: x.name,
        )
        controls = sorted(
            [node for node in self.graph.nodes if isinstance(node, Control)],
            key=lambda x: x.name,  # type: ignore
        )
        ahus = sorted(
            [node for node in self.graph.nodes if isinstance(node, AirHandlingUnit)],
            key=lambda x: x.name,  # type: ignore
        )
        data_bus = DataBus(
            name="data_bus",
            spaces=[space.name for space in spaces],
            external_data=self.external_data,
        )
        self.add_node(data_bus)
        for space in spaces:
            self.graph.add_edge(space, data_bus)
            if space.occupancy:
                self.graph.add_edge(space.occupancy, data_bus)
        for control in controls:
            self.graph.add_edge(control, data_bus)
        for ahu in ahus:
            self.graph.add_edge(ahu, data_bus)
        return data_bus

    def _build_full_space_control(self) -> None:
        spaces = [node for node in self.graph.nodes if isinstance(node, Space)]

        for space in spaces:
            _neighbors = []
            if space.get_last_ventilation_inlet():
                _neighbors += list(
                    self.graph.predecessors(space.get_last_ventilation_inlet())  # type: ignore
                )
            if space.get_last_ventilation_outlet():
                _neighbors += list(
                    self.graph.predecessors(space.get_last_ventilation_outlet())  # type: ignore
                )
            neighbors = list(set(_neighbors))
            controllable_ventilation_elements = list(
                filter(
                    None,
                    [
                        _get_controllable_element(space.ventilation_inlets),
                        _get_controllable_element(space.ventilation_outlets),
                    ],
                )
            )
            for controllable_element in controllable_ventilation_elements:
                if controllable_element.control and isinstance(
                    controllable_element.control, VAVControl
                ):
                    controllable_element.control.ahu = next(
                        (n for n in neighbors if isinstance(n, AirHandlingUnit)), None
                    )

    def _build_occupancy(self, space: "Space") -> None:
        if space.occupancy:
            self.add_node(space.occupancy)
            self.connect_system(space, space.occupancy)

    def connect_spaces(
        self,
        space_1: "Space",
        space_2: "Space",
        internal_element: Optional[
            "InternalElement"
        ] = None,  # TODO: this should not be optional
    ) -> None:
        internal_element = internal_element or InternalElement(
            name=f"internal_{space_1.name}_{space_2.name}",
            surface=10,
            azimuth=10,
            construction=Constructions.internal_wall,
            tilt=Tilt.wall,
        )
        if space_1.position is None or space_2.position is None:
            raise Exception("Position not assigned to spaces")
        internal_element.position = [
            space_1.position[0] + (space_2.position[0] - space_1.position[0]) / 2,
            space_1.position[1],
        ]  # TODO: this is to be moved somewher
        self.add_node(internal_element)
        self.graph.add_edge(
            space_1,
            internal_element,
        )
        self.graph.add_edge(
            space_2,
            internal_element,
        )
        space_1.internal_elements.append(internal_element)
        space_2.internal_elements.append(internal_element)

    def connect_system(self, space: "Space", system: "System") -> None:
        self.graph.add_edge(
            space,
            system,
        )

    def _assign_position(
        self, system_1: System, system_2: System  # :  PLR6301
    ) -> None:
        # TODO: change position to object
        if system_1.position and not system_2.position:
            system_2.position = [system_1.position[0] + 100, system_1.position[1] - 100]
            if hasattr(system_2, "control") and system_2.control:
                system_2.control.position = [
                    system_2.position[0] - 50,
                    system_2.position[1],
                ]
        if system_2.position and not system_1.position:
            system_1.position = [system_2.position[0] - 100, system_2.position[1] - 100]
            if hasattr(system_1, "control") and system_1.control:
                system_1.control.position = [
                    system_1.position[0] - 50,
                    system_1.position[1],
                ]

    def connect_elements(self, element_1: BaseElement, element_2: BaseElement) -> None:
        for element in [element_1, element_2]:
            if element not in self.graph.nodes:
                self.add_node(element)
        self.graph.add_edge(element_1, element_2)
        self._assign_position(element_1, element_2)  # type: ignore

    def connect_systems(self, system_1: System, system_2: System) -> None:

        if system_1 not in self.graph.nodes:
            self.add_node(system_1)
            if system_1.control:
                if system_1.control not in self.graph.nodes:
                    self.add_node(system_1.control)
                    self._system_controls.append(system_1.control)
                self.graph.add_edge(system_1, system_1.control)
                # TODO: check if it is controllable the system

        if system_2 not in self.graph.nodes:
            self.add_node(system_2)
            if system_2.control:
                if system_2.control not in self.graph.nodes:
                    self.add_node(system_2.control)
                    self._system_controls.append(system_2.control)
                self.graph.add_edge(system_2, system_2.control)
        if (
            isinstance(system_2, ThreeWayValve)
            and isinstance(system_1, TemperatureSensor)
        ) or (
            isinstance(system_1, ThreeWayValve)
            and isinstance(system_2, TemperatureSensor)
        ):
            if system_2.control:
                self.graph.add_edge(system_2.control, system_1)
            if system_1.control:
                self.graph.add_edge(system_1.control, system_2)
        self.graph.add_edge(system_1, system_2)
        self._assign_position(system_1, system_2)

    def connect_edges(
        self, edge: Tuple[BaseElement, BaseElement]  # :  PLR6301
    ) -> list[Connection]:
        return connect(edge)

    def merge_spaces(self, space_1: "Space", space_2: "Space") -> None:
        internal_elements = nx.shortest_path(self.graph, space_1, space_2)[1:-1]
        merged_space = space_1 + space_2
        merged_space.internal_elements = internal_elements
        self.graph = nx.contracted_nodes(self.graph, merged_space, space_2)

    def generate_layout(self) -> Dict[Any, Any]:
        # nodes = [n for n in self.graph.nodes if isinstance(n, Space)] # noqa : E800
        # for i, n in enumerate(nodes): # : E800
        #     n.assign_position([200*i, 50]) # noqa : E800

        return nx.spring_layout(self.graph, k=10, dim=2, scale=200)  # type: ignore

    def generate_graphs(self) -> None:
        layout = self.generate_layout()
        for node in self.graph.nodes:
            node.get_position(layout)
            if isinstance(node, Space):
                node.get_neighhors(self.graph)
        # Sorting is necessary here since we need to keep the same
        # index for the same space indatabus
        # TODO: not sure where to put this!!!!
        data_bus = next(
            bus for bus in list(self.graph.nodes) if isinstance(bus, DataBus)
        )
        new_edges = [edge for edge in self.graph.edges if data_bus not in edge]
        edge_with_databus = [edge for edge in self.graph.edges if data_bus in edge]
        edges_with_bus_without_space = [
            edge
            for edge in edge_with_databus
            if not any(isinstance(e, Space) for e in edge)
        ]
        edges_with_bus_with_space = sorted(
            [
                edge
                for edge in edge_with_databus
                if any(isinstance(e, Space) for e in edge)
            ],
            key=lambda e_: next(e for e in e_ if isinstance(e, Space)).name,
        )
        # Sorting is necessary here since we need to keep the
        # same index for the same space indatabus
        # TODO: not sure where to put this!!!!
        for edge in (
            new_edges + edges_with_bus_without_space + edges_with_bus_with_space
        ):
            self.edge_attributes += self.connect_edges(edge)

    def _connect_space_controls(self) -> None:
        undirected_graph = self.graph.to_undirected()
        space_controls = [
            node for node in undirected_graph.nodes if isinstance(node, Space)
        ]
        for space_control in space_controls:
            for system_control in self._system_controls:
                shortest_path(undirected_graph, system_control, space_control)

    def get_ahu_space_elements(self, ahu: AirHandlingUnit) -> List[Space]:
        return [x for x in self._get_ahu_elements(ahu, Space) if isinstance(x, Space)]

    def get_ahu_vav_elements(self, ahu: AirHandlingUnit) -> List[VAV]:
        return [x for x in self._get_ahu_elements(ahu, VAV) if isinstance(x, VAV)]

    def _get_ahu_elements(
        self, ahu: AirHandlingUnit, element_type: Type[Union[VAV, Space]]
    ) -> List[Union[VAV, Space]]:
        elements_: List[Union[VAV, Space]] = []
        elements = [node for node in self.graph.nodes if isinstance(node, element_type)]
        for element in elements:
            try:
                paths = nx.shortest_path(self.graph, ahu, element)
            except Exception as e:
                raise WrongSystemFlowError(
                    "Wrong AHU system configuration flow."
                ) from e
            p = paths[1:-1]
            if p and all(isinstance(p_, Ventilation) for p_ in p):
                elements_.append(element)
        return elements_

    def configure_ahu_control(self) -> None:
        ahus = [node for node in self.graph.nodes if isinstance(node, AirHandlingUnit)]
        for ahu in ahus:
            if ahu.control and isinstance(ahu.control, AhuControl):
                ahu.control.spaces = self.get_ahu_space_elements(ahu)
                ahu.control.vavs = self.get_ahu_vav_elements(ahu)

    def get_linked_valves(self, pump_collector: BaseElement) -> List[Valve]:
        valves_: List[Valve] = []
        valves = [node for node in self.graph.nodes if isinstance(node, Valve)]
        for valve in valves:
            paths = list(nx.all_simple_paths(self.graph, pump_collector, valve))
            for path in paths:
                p = path[1:-1]
                if p and all(isinstance(p_, System) for p_ in p):
                    valves_.append(valve)
                    break
        return valves_

    def configure_collector_control(self) -> None:
        pump_collectors = [
            node
            for node in self.graph.nodes
            if isinstance(node, (Pump, Boiler))
            and isinstance(node.control, CollectorControl)
        ]
        for pump_collector in pump_collectors:
            if isinstance(pump_collector.control, CollectorControl):
                pump_collector.control.valves = self.get_linked_valves(pump_collector)

    def set_weather_path_to_container_path(self, project_path: Path) -> None:
        for node in self.graph.nodes:
            if (
                isinstance(node, Weather)
                and hasattr(node.parameters, "path")
                and node.parameters.path is not None  # type: ignore
            ):
                # TODO: type ognore needs to be fixed
                old_path = Path(node.parameters.path).resolve()  # type: ignore
                if not old_path.exists():
                    parents = [Path.cwd(), *Path.cwd().parents]
                    for parent in parents:
                        old_path = next(parent.rglob(old_path.name), None)  # type: ignore
                        if old_path and old_path.exists():
                            break
                    if not old_path or not old_path.exists():
                        raise FileNotFoundError(f"File {old_path} not found")
                new_path = project_path.joinpath(old_path.name)
                shutil.copy(old_path, new_path)
                # TODO: this is not correct
                node.parameters.path = f'"/simulation/{old_path.name}"'  # type: ignore

    def model(self) -> str:
        Space.counter = 0
        self._build_full_space_control()
        data_bus = self._build_data_bus()
        self.configure_ahu_control()
        self.configure_collector_control()

        data_bus.non_connected_ports = get_non_connected_ports(self.graph.nodes)

        self.generate_graphs()

        self._connect_space_controls()

        element_models = self.build_element_models()
        environment = Environment(
            trim_blocks=True,
            lstrip_blocks=True,
            loader=FileSystemLoader(str(Path(__file__).parent.joinpath("templates"))),
            autoescape=True,
        )
        environment.filters["frozenset"] = frozenset
        environment.filters["enumerate"] = enumerate

        template = environment.get_template("base.jinja2")

        data = extract_properties(self.library, self.name, self.graph.nodes)
        diagram_size = self._get_diagram_size()
        return template.render(
            network=self,
            data=data,
            element_models=element_models,
            library=self.library,
            databus=data_bus,
            dynamic_components=self.dynamic_components,
            diagram_size=diagram_size,
        )

    def _get_diagram_size(self) -> str:
        array = np.array([n.position for n in list(self.graph.nodes)]).T
        x = array[0]
        y = array[1]
        return f"{{{{{min(x) - 50},{min(y) - 50}}},{{{max(x) + 50},{max(y) + 50}}}}}"

    def build_dynamic_component_template(self, node: BaseElement) -> None:

        if node.component_template:
            component = node.component_template.render(
                self.name, node, node.processed_parameters(self.library)
            )
            if node.component_template.category:
                self.dynamic_components[node.component_template.category].append(
                    component
                )

    def build_element_models(self) -> List[str]:
        environment = Environment(
            trim_blocks=True,
            lstrip_blocks=True,
            loader=FileSystemLoader(str(Path(__file__).parent.joinpath("templates"))),
            autoescape=True,
        )
        environment.filters["enumerate"] = enumerate
        models = []
        for node in self.graph.nodes:
            if not node.template:
                continue
            environment.globals.update(self.library.functions)
            rtemplate = environment.from_string(
                "{% import 'macros.jinja2' as macros %}"
                + node.template
                + " "
                + node.annotation_template
            )

            if node.component_template:
                self.build_dynamic_component_template(node)
            model = rtemplate.render(
                element=node,
                package_name=self.name,
                library_name=self.library.name,
                parameters=node.processed_parameters(self.library),
            )
            models.append(model)
        return models

    def add_boiler_plate_spaces(
        self,
        spaces: list[Space],
        create_internal: bool = True,
        weather: Optional[Weather] = None,
    ) -> None:
        for space in spaces:
            self.add_space(space)
        if create_internal:
            for combination in itertools.combinations(spaces, 2):
                self.connect_spaces(*combination)
        weather = weather or Weather()
        weather.position = [-100, 200]  # TODO: move somewhere else
        self.add_node(weather)
        for space in spaces:
            self.connect_system(space, weather)

    def plot(self, use_pyvis: bool = True) -> None:
        if use_pyvis:
            net = PyvisNetwork(notebook=True)
            plot_graph = DiGraph()
            for node in self.graph.nodes:
                plot_graph.add_node(node.name)
            for edge in self.graph.edges:
                plot_graph.add_edge(edge[0].name, edge[1].name)
            net.from_nx(plot_graph)
            net.toggle_physics(True)
            net.show("example.html")
        else:
            nx.draw(self.graph)
            plt.draw()
            plt.show()


def get_non_connected_ports(nodes: List[NodeView]) -> List[BaseInputOutput]:
    port_types = ["Real", "Integer", "Boolean"]
    ports: Dict[str, List[BaseInputOutput]] = {
        f"{port_type}{direction}": []
        for port_type in port_types
        for direction in ["Output", "Input"]
    }

    for node in nodes:
        if not (
            hasattr(node, "component_template")
            and hasattr(node.component_template, "bus")
        ):
            continue
        if node.component_template and node.component_template.bus:
            node_ports = node.component_template.bus.list_ports(node)
            for port_type in port_types:
                ports[f"{port_type}Output"] += node_ports[f"{port_type}Output"]
                ports[f"{port_type}Input"] += node_ports[f"{port_type}Input"]

    for port_type in port_types:
        ports[f"{port_type}Output"] = list(set(ports[f"{port_type}Output"]))
        ports[f"{port_type}Input"] = list(set(ports[f"{port_type}Input"]))

    return list(
        itertools.chain(
            *[
                _get_non_connected_ports_intersection(
                    ports[f"{port_type}Input"], ports[f"{port_type}Output"]
                )
                for port_type in port_types
            ]
        )
    )


def _get_non_connected_ports_intersection(
    input_ports: List[BaseInputOutput], output_ports: List[BaseInputOutput]
) -> List[BaseInputOutput]:
    return list(set(input_ports) - set(output_ports).intersection(set(input_ports)))



import re
from typing import TYPE_CHECKING, Any, Callable, List

from trano import elements

if TYPE_CHECKING:

    from trano.elements import Port


def to_snake_case(name: str) -> str:
    s1 = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", name)
    return re.sub("([a-z0-9])([A-Z])", r"\1_\2", s1).lower()


def compose_func(ports_: List["Port"]) -> Callable[[], List["Port"]]:
    return lambda: ports_


def _get_type(_type: Any) -> Any:  # noqa: ANN401
    if _type == "string":
        return str
    elif _type == "float":
        return float
    elif _type == "integer":
        return int
    elif _type == "boolean":
        return bool
    else:
        raise Exception("Unknown type")


def _get_default(v: Any) -> Any:  # noqa: ANN401
    if "ifabsent" not in v:
        return None
    tag = v["range"]
    if tag == "integer":
        tag = "int"
    value = v["ifabsent"].replace(tag, "")[1:-1]
    if value == "None":
        return None

    try:
        return _get_type(v["range"])(value)
    except Exception as e:

        raise e


# TODO: class names should be standardized!!
def import_element_function(function_name: str) -> Any:  # noqa: ANN401
    attribute = [
        attribute
        for attribute in elements.__all__
        if attribute.lower() == function_name.lower()
    ]
    if len(attribute) > 1:
        raise Exception(f"Element {function_name} has more than one match")
    if len(attribute) == 0:
        raise Exception(f"Element {function_name} not found")
    return getattr(elements, attribute[0])



import abc
from typing import TYPE_CHECKING, List, Optional

from pydantic import computed_field

from trano.elements import BaseElement, BaseSimpleWall, BaseWall
from trano.elements.envelope import MergedBaseWall
from trano.elements.space import Space
from trano.elements.system import Boiler, System
from trano.reporting.types import (
    BaseTable,
    ConstructionTable,
    ContentDocumentation,
    ResultFile,
    SpaceTable,
    SystemTable,
    Topic,
)
from trano.reporting.utils import _dump_list_attributes, _get_elements, get_description

if TYPE_CHECKING:
    from trano.topology import Network


class BaseDocumentation(ContentDocumentation):
    title: Optional[str] = None
    introduction: Optional[str] = None
    table: BaseTable | List[BaseTable]
    conclusions: Optional[str] = None

    @classmethod
    @abc.abstractmethod
    def from_elements(
        cls, elements: List[BaseElement], content_documentation: ContentDocumentation
    ) -> "BaseDocumentation": ...

    @computed_field
    def topic(self) -> Topic:
        return type(self).__name__.replace("Documentation", "")  # type: ignore


class SpacesDocumentation(BaseDocumentation):
    table: List[SpaceTable]  # type: ignore

    @classmethod
    def from_elements(
        cls, elements: List[BaseElement], content_documentation: ContentDocumentation
    ) -> "SpacesDocumentation":
        spaces = _get_elements(elements, Space)
        spaces_ = []
        boundary_parameters = {
            "name": True,
            "surface": True,
            "type": True,
            "azimuth": True,
            "tilt": True,
            "construction": {"name"},
        }
        system_parameters = {"name": True, "parameters": True, "type": True}
        document_mapping = {
            "emissions": system_parameters,
            "ventilation_inlets": system_parameters,
            "ventilation_outlets": system_parameters,
            "external_boundaries": boundary_parameters,
            "internal_elements": boundary_parameters,
        }
        for space in spaces:
            main_space = space.model_dump(
                mode="json",
                include={  # type: ignore
                    "name": True,
                    "occupancy": {"name": True},
                },
                exclude_none=True,
                by_alias=True,
            )
            if space.parameters:
                main_space["parameters"] = space.parameters.model_dump(
                    by_alias=True, exclude_none=True
                )
            if (
                main_space.get("occupancy")
                and isinstance(space, Space)
                and space.occupancy
                and space.occupancy.parameters
            ):
                main_space["occupancy"]["parameters"] = (
                    space.occupancy.parameters.model_dump(
                        by_alias=True, exclude_none=True
                    )
                )
            for key, value in document_mapping.items():
                values = _dump_list_attributes(space, key, value)
                if values:
                    main_space[key] = _dump_list_attributes(space, key, value)
            spaces_.append(SpaceTable(**main_space))
        data = {
            "table": spaces_,
        } | content_documentation.model_dump(exclude_none=True)
        return cls(**data)


class ConstructionDocumentation(BaseDocumentation):
    table: List[ConstructionTable]  # type: ignore

    @classmethod
    def from_elements(
        cls, elements: List[BaseElement], content_documentation: ContentDocumentation
    ) -> "ConstructionDocumentation":
        constructions = [
            c.model_dump(by_alias=True, exclude_none=True)
            for c in {
                construction
                for w in _get_elements(elements, BaseWall)
                if isinstance(w, (BaseSimpleWall, MergedBaseWall))
                for construction in (
                    [w.construction] if hasattr(w, "construction") else w.constructions
                )
            }
        ]
        data = {
            "table": [
                ConstructionTable(**construction) for construction in constructions
            ],
        } | content_documentation.model_dump(exclude_none=True)
        return cls(**data)


class SystemsDocumentation(BaseDocumentation):
    table: List[SystemTable]  # type: ignore

    @classmethod
    def from_elements(
        cls, elements: List[BaseElement], content_documentation: ContentDocumentation
    ) -> "SystemsDocumentation":
        spaces = _get_elements(elements, Space)
        get_description()
        systems_to_exclude = {
            system
            for space in spaces
            if isinstance(space, Space)
            for system in space.emissions
            + space.ventilation_inlets
            + space.ventilation_outlets
        }
        systems = []
        for system in _get_elements(elements, System):
            if system not in systems_to_exclude:
                system_ = system.model_dump(
                    by_alias=True,
                    exclude_none=True,
                    include={"name": True, "parameters": True, "type": True},
                )
                if system.parameters:
                    system_["parameters"] = system.parameters.model_dump()
                systems.append(SystemTable(**system_))

        data = {
            "table": systems,
        } | content_documentation.model_dump(exclude_none=True)
        return cls(**data)


class ContentModelDocumentation(ContentDocumentation):
    spaces: ContentDocumentation
    constructions: ContentDocumentation
    systems: ContentDocumentation


class ModelDocumentation(ContentDocumentation):
    spaces: SpacesDocumentation
    constructions: ConstructionDocumentation
    systems: SystemsDocumentation
    elements: List[BaseElement]
    result: Optional[ResultFile] = None

    @classmethod
    def from_model_elements(
        cls,
        elements: List[BaseElement],
        content_documentation: ContentModelDocumentation,
        result: Optional[ResultFile] = None,
    ) -> "ModelDocumentation":
        spaces_documentation = SpacesDocumentation.from_elements(
            elements, content_documentation.spaces
        )
        constructions = ConstructionDocumentation.from_elements(
            elements, content_documentation.constructions
        )
        systems = SystemsDocumentation.from_elements(
            elements, content_documentation.systems
        )
        data = content_documentation.model_dump(exclude_none=True) | {
            "spaces": spaces_documentation,
            "constructions": constructions,
            "systems": systems,
            "elements": elements,
            "result": result,
        }

        return cls(**data)

    @classmethod
    def from_network(
        cls,
        network: "Network",
        content_model_documentation: Optional[ContentModelDocumentation] = None,
        result: Optional[ResultFile] = None,
    ) -> "ModelDocumentation":
        content_model_documentation = (
            content_model_documentation
            or ContentModelDocumentation(
                spaces=ContentDocumentation(),
                constructions=ContentDocumentation(),
                systems=ContentDocumentation(),
            )
        )
        elements = [
            x
            for x in list(network.graph.nodes)
            if isinstance(x, (Boiler, Space, BaseSimpleWall, MergedBaseWall))
        ]
        return cls.from_model_elements(elements, content_model_documentation, result)  # type: ignore



from copy import deepcopy
from functools import partial
from pathlib import Path
from typing import TYPE_CHECKING, Any, Callable, Dict, List, Optional

import yaml
from pydantic import BaseModel, ConfigDict, Field, create_model

from trano.elements.components import COMPONENTS, DynamicComponentTemplate
from trano.elements.parameters import default_parameters
from trano.elements.types import BaseVariant
from trano.elements.utils import compose_func

if TYPE_CHECKING:
    from trano.elements import BaseParameter, Figure, Port, WallParameters


def tilts_processing_ideas(element: "WallParameters") -> List[str]:
    return [f"IDEAS.Types.Tilt.{tilt.value.capitalize()}" for tilt in element.tilts]


class Templates(BaseModel):
    is_package: bool = False
    construction: str
    glazing: str
    material: Optional[str] = None
    main: str


def read_libraries() -> Dict[str, Dict[str, Any]]:
    library_path = Path(__file__).parent.joinpath("library.yaml")
    data: Dict[str, Dict[str, Any]] = yaml.safe_load(library_path.read_text())
    return data


class Library(BaseModel):
    name: str
    merged_external_boundaries: bool = False
    functions: Dict[str, Callable[[Any], Any]] = {
        "tilts_processing_ideas": tilts_processing_ideas
    }
    constants: str = ""
    templates: Templates
    default: bool = False
    default_parameters: Dict[str, Any] = Field(
        default_factory=dict
    )  # TODO: this should be baseparameters

    @classmethod
    def from_configuration(cls, name: str) -> "Library":
        libraries = read_libraries()

        if name not in libraries:
            raise ValueError(
                f"Library {name} not found. Available libraries: {list(libraries)}"
            )
        library_data = libraries[name]
        return cls(**library_data)

    @classmethod
    def load_default(cls) -> "Library":
        libraries = read_libraries()
        default_library = [
            library_data
            for _, library_data in libraries.items()
            if library_data.get("default")
        ]
        if not default_library:
            raise ValueError("No default library found")
        return cls(**default_library[0])


class AvailableLibraries(BaseModel):
    ideas: List[Callable[[], "LibraryData"]] = Field(default=[lambda: None])
    buildings: List[Callable[[], "LibraryData"]] = Field(default=[lambda: None])

    def get_library_data(self, library: "Library", variant: str) -> Any:  # noqa: ANN401
        selected_variant = [
            variant_
            for variant_ in getattr(self, library.name.lower())
            if variant_().variant == variant
        ]
        if not selected_variant:
            return
        # TODO: to be more strict
        return selected_variant[0]()

    # TODO: this code should be library independent
    @classmethod
    def from_config(cls, name: str) -> Optional["AvailableLibraries"]:
        from trano.elements import BaseParameter, Figure, Port, parameters

        data = deepcopy(COMPONENTS)
        components_data__ = [
            component
            for component in data["components"]
            for classes_ in component["classes"]
            if name == classes_
        ]
        if not components_data__:
            return None
        components: Dict[str, Any] = {"ideas": [], "buildings": []}
        for component in components_data__:
            dynamic_component = None
            if component.get("component_template"):
                dynamic_component = DynamicComponentTemplate(
                    **component["component_template"]
                )
            if component["parameter_processing"].get("parameter", None):
                function_name = component["parameter_processing"]["function"]
                parameter_processing = partial(
                    getattr(parameters, function_name),
                    **{
                        function_name: component["parameter_processing"].get(
                            "parameter", {}
                        )
                    },
                )
            else:
                parameter_processing = getattr(
                    parameters, component["parameter_processing"]["function"]
                )

            component_ = create_model(
                f"Base{component['library'].capitalize() }{name.capitalize()}",
                __base__=LibraryData,
                template=(str, f"{component['template']}"),
                ports_factory=(
                    Callable[[], List[Port]],
                    compose_func([Port(**port) for port in component["ports"]]),
                ),
                component_template=(DynamicComponentTemplate, dynamic_component),
                variant=(str, component["variant"]),
                figures=(
                    List[Figure],
                    Field([Figure(**fig) for fig in component.get("figures", [])]),
                ),
                parameter_processing=(
                    Callable[[BaseParameter], Dict[str, Any]],
                    parameter_processing,
                ),
            )

            if component["library"] == "default":
                components["ideas"].append(component_)
                components["buildings"].append(component_)
            else:
                components[component["library"]].append(component_)
        return cls(**components)


class LibraryData(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)
    template: str = ""
    component_template: Optional[DynamicComponentTemplate] = None
    ports_factory: Callable[[], List["Port"]]
    variant: str = BaseVariant.default
    parameter_processing: Callable[["BaseParameter"], Dict[str, Any]] = (
        default_parameters
    )
    figures: List["Figure"] = Field(default=[])



from typing import List, Optional

from pydantic import computed_field

from trano.elements import Control
from trano.elements.base import BaseElement
from trano.elements.parameters import BaseParameter
from trano.elements.space import Space
from trano.elements.system import VAV, Valve


class ControlParameters(BaseParameter): ...


class VAVControl(Control):
    ahu: Optional[BaseElement] = None


class ThreeWayValveControl(Control): ...


class EmissionControl(Control): ...


class CollectorControl(Control):
    valves: Optional[List[Valve]] = None


class BoilerControl(Control): ...


class AhuControl(Control):
    spaces: Optional[List[Space]] = None
    vavs: Optional[List[VAV]] = None

    @computed_field
    def zon_gro_mat(self) -> str:
        if self.vavs is None:
            return "[1]"
        return str([1] * len(self.vavs))

    @computed_field
    def zon_gro_mat_tra(self) -> str:
        if self.vavs is None:
            return "[1]"
        return str([1] * len(self.vavs)).replace(",", ";")



import webbrowser
from pathlib import Path

from trano.data_models.conversion import convert_network
from trano.library.library import Library
from trano.reporting.html import to_html_reporting
from trano.reporting.reporting import ModelDocumentation
from trano.reporting.types import ResultFile
from trano.simulate.simulate import SimulationLibraryOptions, simulate
from trano.topology import Network
from trano.utils.utils import is_success


def _create_network(model: Path | str, library: str) -> Network:
    library_ = Library.from_configuration(library)
    model = Path(model).resolve()
    return convert_network(model.stem, model, library=library_)


def create_model(model: Path | str, library: str = "Buildings") -> None:
    network = _create_network(model, library)
    modelica_model = network.model()
    Path(model).resolve().with_suffix(".mo").write_text(modelica_model)


def simulate_model(model: Path | str, options: SimulationLibraryOptions) -> None:

    network = _create_network(model, options.library_name)
    model = Path(model).resolve()
    results = simulate(
        model.parent,
        network,
        options=options,
    )
    if not is_success(results):
        raise ValueError("Simulation failed")

    reporting = ModelDocumentation.from_network(
        network,
        result=ResultFile(
            path=Path(model.parent) / "results" / f"{model.stem}.building_res.mat"
        ),
    )
    html = to_html_reporting(reporting)
    report_path = Path(model.parent / f"{model.stem}.html")
    report_path.write_text(html)
    webbrowser.open(f"file://{report_path}")


def report(model: Path | str, options: SimulationLibraryOptions) -> None:
    model = Path(model).resolve()
    network = convert_network(
        model.stem, model, library=Library.from_configuration(options.library_name)
    )
    reporting = ModelDocumentation.from_network(
        network,
        result=ResultFile(
            path=Path(model.parent) / "results" / f"{model.stem}.building_res.mat"
        ),
    )
    html = to_html_reporting(reporting)
    report_path = Path(model.parent / f"{model.stem}.html")
    report_path.write_text(html)
    webbrowser.open(f"file://{report_path}")
